/*
 * FORM METADATA UTILITIES
 *
 * Exports
 * -------
 * autoMeta(model) - generate basic metadata from a raw or mapped model, recursively if necessary
 * combineExtraMeta(metadata, extraMeta) - combine extra metadata into metatdata, lazyly and recursively
 * combineModelWithMeta(model, extraMeta) - automatically generate metadata from model then combine extra metadata
 * buildFieldSpecificMeta(metadata) - use field-type-specific metadata models to expand metadata
 * buildFormGroupFunctionFactory(fb) - return a function to buildFormGroups using the supplied FormBuilder singleton
 *
 * Variable names
 * --------------
 * metaF   = metadata for Field
 * metaG   = metadata for Group (possibly nested)
 * metaFoG = metadata for Field Or Group
 *
 */

import { FormBuilder, FormGroup, FormArray, FormControl, ValidatorFn, AsyncValidatorFn } from '@angular/forms';
import { reduce } from 'lodash/fp';
import * as fmdModels from '../models/field.model';

// REMINDER: Import this directly from @angular/forms when we upgrade to Angular 6
// While we're still on Angular 5 just declare it
interface AbstractControlOptions {
	validators?: ValidatorFn | ValidatorFn[] | null;
	asyncValidators?: AsyncValidatorFn | AsyncValidatorFn[] | null;
	updateOn?: 'change' | 'blur' | 'submit';
}


// ---------------------------------------------------------------------------------------------------------------------
// AutoMeta: Generate Autimatic Metadata from a raw model (or mapped raw model)
// ---------------------------------------------------------------------------------------------------------------------

const isScalar = val => typeof val === 'boolean' || typeof val === 'number' || typeof val === 'string';
const isArray = val => Array.isArray(val);

const keyValPairToMeta = (val, key) => ({ name: key, [isScalar(val) ? 'value' : 'meta']: val });
const keyValPairToMetaRecursive = ( [key, val] ) => {
	const innerVal = isScalar(val) ? val : (isArray(val) ? arrayToMeta(val) : autoMeta(val));
	const metaVal = keyValPairToMeta(innerVal, key);
	return [key, metaVal];
};
const arrayToMeta = array => array.map(val => ({ name: val, 'value' : val }));

const autoMeta = model => Object.entries(model)
	.map(keyValPairToMetaRecursive)
	.reduce((res, [key, val]) => addProp(res, key, val), {});


// ---------------------------------------------------------------------------------------------------------------------
// Combine automatically generated metadata with overrides
// ---------------------------------------------------------------------------------------------------------------------

const combineMetaForField = (metaF, extraMetaF) => Object.assign(metaF, extraMetaF);
const combineExtraMeta = (metaG, extraMeta, createFromExtra = false) => {
	const combinedMeta = {};
	Object.entries(extraMeta).forEach(([key, val]) => {
		if (typeof metaG[key] === 'object' || createFromExtra) {
			combinedMeta[key] = metaG[key] && (<any>val).meta ?
				combineMetaForField(metaG[key], { meta: combineExtraMeta(metaG[key].meta, (<any>val).meta, createFromExtra) }) :
				combineMetaForField(metaG[key] || {}, val);
		}
	});
	return { ...metaG, ...combinedMeta };
};

// Combine model with overrides (after automatically generating metadata from the model)
const combineModelWithMeta = (model, extraMeta, createFromExtra = false) => combineExtraMeta(autoMeta(model), extraMeta, createFromExtra);


// ---------------------------------------------------------------------------------------------------------------------
// Build Form-Field-Type-Specific Metadata (using the field models in dynaform/models)
// ---------------------------------------------------------------------------------------------------------------------

const buildFieldClassName = (t = 'text') => {
	const start = t[0].toUpperCase() + t.slice(1);
	if (start === 'Container' || t.slice(-5) === 'Group') {
		return start;
	}
	return start + 'Field';
};

const buildModeledField = metaFoG => {
	const type = isContainer(metaFoG) ? 'container' : metaFoG.type;
	const className = buildFieldClassName(type);
	if (!fmdModels[className]) {
		throw new Error(`No metadata model "${className}" for type "${type}"`);
	}
	return new fmdModels[className](metaFoG);
};

// Build Form Group Member
const buildModeledFieldGroupMember = metaFoG => {
	const modeledGroupMember = buildModeledField(metaFoG);
	if (isContainer(metaFoG)) {
		modeledGroupMember.meta = _buildFieldSpecificMeta(metaFoG.meta);
	}
	return modeledGroupMember;
};

// Build Form Group
const buildModeledFieldGroupReducerIteree = (res, metaFoG) => Object.assign(res, { [metaFoG.name]: buildModeledFieldGroupMember(metaFoG) });
const _buildFieldSpecificMeta = metaG => reduce(buildModeledFieldGroupReducerIteree, {}, metaG);
const buildFieldSpecificMeta = metaG => _buildFieldSpecificMeta(addMissingNames(metaG));


// ---------------------------------------------------------------------------------------------------------------------
// Build FormGroups
// ---------------------------------------------------------------------------------------------------------------------

const buildFormGroupFunctionFactory = (fb: FormBuilder): (meta) => FormGroup => {
	// Establishes a closre over the supplied FormBuilder and returns a function that builds FormGroups from metadata
	// ( it's done this way so we can use the FormBuilder singleton without reinitialising )

	// Build Form Control
	const buildControlState = metaF => ({ value: metaF.value || metaF.default, disabled: metaF.isDisabled });
	const buildValidators = (metaF): AbstractControlOptions => ({
		validators: metaF.validators,
		asyncValidators: metaF.asyncValidators,
		updateOn: 'change' // blur not working for custom components - maybe use change for custom and blur for text
	});
	const buildFormControl = metaF => new FormControl(buildControlState(metaF), buildValidators(metaF));

	// Build Form Array
	const buildFormArray = (metaG): FormArray => fb.array(metaG.map(metaF => buildFormControl(metaF)));

	// Build Form Group Member - builds a FormControl, FormArray, or another FormGroup which can contain any of these
	const buildFormGroupMember = metaFoG => isGroup(metaFoG) ?
		(isArray(metaFoG.meta) ? buildFormArray(metaFoG.meta) : _buildFormGroup(metaFoG.meta)) :
		buildFormControl(metaFoG);

	const buildFormGroupReducerIteree = (res, metaFoG) => Object.assign(res, { [metaFoG.name]: buildFormGroupMember(metaFoG) });
	const _buildFormGroup = _metaG => fb.group(reduce(buildFormGroupReducerIteree, {}, _metaG));

	// The main function - builds FormGroups containing other FormGroups, FormArrays and FormControls
	const buildFormGroup = metaG => {
		// Ensure that we have Field-Specific Metadata, not raw Objects
		const metaWithNameKeys = addMissingNames(metaG);
		// MAYBE only run this if first entry isn't right, for reasons of efficiency
		const fieldModeledMeta = addMissingFieldSpecificMeta(metaWithNameKeys);
		return _buildFormGroup(fieldModeledMeta);
	};
	return buildFormGroup;
};


// ---------------------------------------------------------------------------------------------------------------------
// Helper Functions
// ---------------------------------------------------------------------------------------------------------------------

// Add Property to object
const addProp = (obj, key, val) => { obj[key] = val; return obj; };

// Is Group
// Helper function to distinguish group from field
const isGroup = (metaFoG): boolean => !!metaFoG.meta;

// Is Container
// Helper function to distinguish container group (a group of child fields)
const isContainer = (metaFoG): boolean => isGroup(metaFoG) && (!metaFoG.type || metaFoG.type.toLowerCase() === 'container');

// Add Missing Names
// Helper function to add any missing 'name' properties to Fields and Groups using property's key, recursively
const addNameIfMissing = (metaFoG, key) => metaFoG.name ? metaFoG : addProp(metaFoG, 'name', key);
const addNameToSelfAndChildren = ( [key, metaFoG] ) => {
	metaFoG = addNameIfMissing(metaFoG, key);
	if (isGroup(metaFoG)) {
		metaFoG.meta = isArray(metaFoG.meta) ? Object.values(addMissingNames(metaFoG.meta)) : addMissingNames(metaFoG.meta); // Recursion
	}
	return [key, metaFoG];
};
const addMissingNames = metaG => Object.entries(metaG)
	.map(addNameToSelfAndChildren)
	.reduce((res, [key, val]) => addProp(res, key, val), {});


// Add Missing Field-Specific Meta
// Helper function to add any missing Field-Specific Metadata  (using models in dynaform/models), recursively
// Checks the constrctor, which should NOT be a plain Object, but rather TextField, TextareaField, SelectField, etc.
const add_FSM_IfMissing = metaFoG => metaFoG.constructor.name === 'Object' ? buildModeledFieldGroupMember(metaFoG) : metaFoG;
const add_FSM_ToSelfAndChildren = ( [key, metaFoG] ) => {
	metaFoG = add_FSM_IfMissing(metaFoG);
	return [key, metaFoG];
};
const addMissingFieldSpecificMeta = metaG => Object.entries(metaG)
	.map(add_FSM_ToSelfAndChildren)
	.reduce((res, [key, val]) => addProp(res, key, val), {});


// ---------------------------------------------------------------------------------------------------------------------
// Exports
// ---------------------------------------------------------------------------------------------------------------------

export { autoMeta, combineModelWithMeta, combineExtraMeta, buildFieldSpecificMeta, buildFormGroupFunctionFactory };
